ParseNode subclass: #MethodNode
	instanceVariableNames: 'selectorOrFalse precedence arguments block literals primitive encoder temporaries '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Compiler'!
!MethodNode commentStamp: '<historical>' prior: 0!
I am the root of the parse tree.!


!MethodNode methodsFor: 'initialize-release' stamp: 'tk 8/3/1999 12:47'!
block
	^ block! !

!MethodNode methodsFor: 'initialize-release'!
selector: selOrFalse arguments: args precedence: p temporaries: temps block: blk encoder: anEncoder primitive: prim 
	"Initialize the receiver with respect to the arguments given."

	encoder _ anEncoder.
	selectorOrFalse _ selOrFalse.
	precedence _ p.
	arguments _ args.
	temporaries _ temps.
	block _ blk.
	primitive _ prim! !


!MethodNode methodsFor: 'code generation'!
encoder
	^ encoder! !

!MethodNode methodsFor: 'code generation' stamp: 'di 5/25/2000 06:48'!
generate: trailer 
	"The receiver is the root of a parse tree. Answer a CompiledMethod. The
	argument, trailer, is the references to the source code that is stored with 
	every CompiledMethod."

	| blkSize nLits stack strm nArgs method |
	self generate: trailer ifQuick: 
		[:m |  method _ m.
		method cacheTempNames: self tempNames.
		^ method].
	nArgs _ arguments size.
	blkSize _ block sizeForEvaluatedValue: encoder.
	literals _ encoder allLiterals.
	(nLits _ literals size) > 255
		ifTrue: [^self error: 'Too many literals referenced'].
	method _ CompiledMethod	"Dummy to allocate right size"
				newBytes: blkSize
				trailerBytes: trailer 
				nArgs: nArgs
				nTemps: encoder maxTemp
				nStack: 0
				nLits: nLits
				primitive: primitive.
	strm _ ReadWriteStream with: method.
	strm position: method initialPC - 1.
	stack _ ParseStack new init.
	block emitForEvaluatedValue: stack on: strm.
	stack position ~= 1 ifTrue: [^self error: 'Compiler stack
discrepancy'].
	strm position ~= (method size - trailer size) 
		ifTrue: [^self error: 'Compiler code size discrepancy'].
	method needsFrameSize: stack size.
	1 to: nLits do: [:lit | method literalAt: lit put: (literals at: lit)].
	method cacheTempNames: self tempNames.
	^ method! !

!MethodNode methodsFor: 'code generation' stamp: 'di 5/25/2000 06:45'!
generate: trailer ifQuick: methodBlock
	| v |
	(primitive = 0 and: [arguments size = 0 and: [block isQuick]])
		ifFalse: [^ self].
	v _ block code.
	v < 0
		ifTrue: [^ self].
	v = LdSelf
		ifTrue: [^ methodBlock value: (CompiledMethod toReturnSelfTrailerBytes: trailer)].
	(v between: LdTrue and: LdMinus1 + 3)
		ifTrue: [^ methodBlock value: (CompiledMethod toReturnConstant: v - LdSelf trailerBytes: trailer)].
	v < ((CodeBases at: LdInstType) + (CodeLimits at: LdInstType))
		ifTrue: [^ methodBlock value: (CompiledMethod toReturnField: v trailerBytes: trailer)].
	v // 256 = 1
		ifTrue: [^ methodBlock value: (CompiledMethod toReturnField: v \\ 256 trailerBytes: trailer)]! !

!MethodNode methodsFor: 'code generation'!
selector 
	"Answer the message selector for the method represented by the receiver."

	(selectorOrFalse isMemberOf: Symbol)
		ifTrue: [^selectorOrFalse].
	^selectorOrFalse key! !

!MethodNode methodsFor: 'code generation'!
sourceMap
	"Answer a SortedCollection of associations of the form: pc (byte offset in 
	me) -> sourceRange (an Interval) in source text."

	self generate: #(0 0 0 0).
	^encoder sourceMap! !


!MethodNode methodsFor: 'converting'!
decompileString 
	"Answer a string description of the parse tree whose root is the receiver."
	^ String streamContents: [:strm | self printOn: strm]! !

!MethodNode methodsFor: 'converting'!
decompileText 
	"Answer a string description of the parse tree whose root is the receiver."
	^ Text streamContents: [:strm | self printOn: strm]! !


!MethodNode methodsFor: 'printing' stamp: 'sw 11/17/1999 13:57'!
printOn: aStream 
	| args |
	precedence = 1
		ifTrue: 
			[aStream nextPutAll: self selector]
		ifFalse: 
			[args _ ReadStream on: arguments.
			self selector keywords do: 
				[:s | 
				aStream nextPutAll: s; space.
				aStream withAttributes: (Preferences syntaxAttributesFor: #methodArgument)
					do: [aStream nextPutAll: args next key].
				aStream space]].
	comment == nil ifFalse: 
			[aStream crtab: 1.
			self printCommentOn: aStream indent: 1].
	temporaries size > 0 ifTrue: 
			[aStream crtab: 1.
			aStream nextPutAll: '| '.
			aStream withAttributes: (Preferences syntaxAttributesFor: #temporaryVariable)
				do:
					[temporaries do: 
						[:temp | 
						aStream nextPutAll: temp key.
						aStream space]].
			aStream nextPut: $|].
	primitive > 0 ifTrue:
			[(primitive between: 255 and: 519) ifFalse:  " Dont decompile <prim> for, eg, ^ self "
				[aStream crtab: 1.
				self printPrimitiveOn: aStream]].
	aStream crtab: 1.
	^ block printStatementsOn: aStream indent: 0! !

!MethodNode methodsFor: 'printing' stamp: 'ar 11/28/1999 19:38'!
printPrimitiveOn: aStream
	"Print the primitive on aStream"
	| primIndex primDecl |
	primIndex _ primitive.
	primIndex = 0 ifTrue:[^self].
	primIndex = 120 "External call spec"
		ifTrue:[^aStream print: encoder literals first].
	aStream nextPutAll: '<primitive: '.
	primIndex = 117 ifTrue:[
		primDecl _ encoder literals at: 1.
		aStream 
			nextPut: $';
			nextPutAll: (primDecl at: 2);
			nextPut:$'.
		(primDecl at: 1) notNil ifTrue:[
			aStream 
				nextPutAll:' module:';
				nextPut:$';
				nextPutAll: (primDecl at: 1);
				nextPut:$'.
		].
	] ifFalse:[aStream print: primIndex].
	aStream nextPut: $>.! !

!MethodNode methodsFor: 'printing'!
tempNames
	^ encoder tempNames! !


!MethodNode methodsFor: 'C translation' stamp: 'jm 11/15/2003 04:56'!
asTMethodFromClass: aClass

	^ TMethod new
		setSelector: selectorOrFalse
		args: arguments
		locals: encoder tempsAndBlockArgs
		block: block
		primitive: primitive
! !


!MethodNode methodsFor: 'elements' stamp: 'jens 6/4/2009 17:09'!
asElementFromText
	| se |
	se _ self asUnnamedMethodElement.
	temporaries isEmpty
		ifTrue: [^se steps]
		ifFalse: [^se]! !

!MethodNode methodsFor: 'elements' stamp: 'jens 2/2/2009 23:18'!
asSyntaxElement
	| element pe be |
	element _ MethodElementMorph new
		selector: self selector asString 
		arguments: (arguments collect: [:arg| arg key]).

	temporaries do: [:temp |
		element addVariableNamed: temp key ].

	primitive > 0 ifTrue:
			[(primitive between: 255 and: 519) ifFalse: 
				[pe _ self primitiveElement.
				element addStep: pe]].

	be _ (block asStackOfSteps).
	be ifNotNil:[	element addStep: be].
	^element! !

!MethodNode methodsFor: 'elements' stamp: 'jens 6/4/2009 17:10'!
asUnnamedMethodElement
	| element pe be |
	element _ MethodElementMorph new.

	temporaries do: [:temp |
		element addVariableNamed: temp key ].

	primitive > 0 ifTrue:
			[(primitive between: 255 and: 519) ifFalse: 
				[pe _ self primitiveElement.
				element addStep: pe]].

	be _ (block asStackOfSteps).
	be ifNotNil:[	element addStep: be].
	^element! !

!MethodNode methodsFor: 'elements' stamp: 'jens 1/19/2009 23:59'!
primitiveElement

	| s primIndex primDecl |
	s _ WriteStream on: String new.
	primIndex _ primitive.

	primIndex = 120 "External call spec"
		ifTrue:[1 halt. ^s print: encoder literals first].
	s nextPutAll: 'primitive: '.
	primIndex = 117 ifTrue:[
		primDecl _ encoder literals at: 1.
		s 
			nextPut: $';
			nextPutAll: (primDecl at: 2);
			nextPut:$'.
		(primDecl at: 1) notNil ifTrue:[
			s 
				nextPutAll:' module:';
				nextPut:$';
				nextPutAll: (primDecl at: 1);
				nextPut:$'.
		].
	] ifFalse:[s print: primIndex].
	^PrimitiveElementMorph new contents: s contents! !
