ImageReadWriter subclass: #PNGReadWriter
	instanceVariableNames: 'chunk form width height depth backColor bitsPerChannel colorType interlaceMethod bitsPerPixel bytesPerScanline thisScanline prevScanline rowSize globalDataChunk unknownChunks palette transparentPixelValue filtersSeen bigEndian '
	classVariableNames: 'BPP BlockHeight BlockWidth Debugging StandardColors StandardSwizzleMaps '
	poolDictionaries: ''
	category: 'Graphics-Files'!
!PNGReadWriter commentStamp: 'jm 10/14/2003 16:53' prior: 0!
I am a subclass of ImageReadWriter that reads Portable Network Graphics (PNG) images.

Submitted by Duane Maxwell.
!


!PNGReadWriter methodsFor: 'accessing' stamp: 'RAA 11/7/2000 09:20'!
debugging

	^ Debugging == true
! !

!PNGReadWriter methodsFor: 'accessing' stamp: 'jens 11/18/2008 20:33'!
nextImage
	bigEndian _ Smalltalk isBigEndian.
	filtersSeen _ Bag new.
	globalDataChunk _ nil.
	transparentPixelValue _ nil.
	unknownChunks _ Set new.
	stream reset.
	stream binary.
	stream skip: 8.

	[stream atEnd] whileFalse: [self processNextChunk].

	"set up the form"
	palette ifNotNil: [
		"dump the palette if it's the same as our standard palette"
		palette = (StandardColors copyFrom: 1 to: palette size) ifTrue: [palette _ nil]].

	(depth <= 8 and: [palette notNil])
		ifTrue: [
			form _ ColorForm extent: width @ height depth: depth.
			form colors: palette]
		ifFalse: [
			form _ Form extent: width @ height depth: depth].

	backColor ifNotNil: [form fillColor: backColor].
	chunk _ globalDataChunk ifNil: [self error: 'image data is missing'].
	chunk ifNotNil: [self processIDATChunk].

	(self debugging and: [unknownChunks isEmpty not]) ifTrue: [
		Transcript cr; show: 'unknown chunks: '.
		Transcript show: unknownChunks asSortedCollection asArray printString].

	self debugging ifTrue: [
		Transcript cr; show: 'form = ' , form printString.
		Transcript cr; show: 'colorType = ' , colorType printString.
		Transcript cr; show: 'interlaceMethod = ' , interlaceMethod printString.
		Transcript cr; show: 'filters = ' , filtersSeen sortedCounts asArray printString].
	^ form
! !

!PNGReadWriter methodsFor: 'accessing' stamp: 'DSM 3/24/2000 01:12'!
understandsImageFormat
	#(137 80 78 71 13 10 26 10) do: [ :byte |
		stream next = byte ifFalse: [^ false]].
	^ true
! !


!PNGReadWriter methodsFor: 'chunks' stamp: 'jm 3/16/2009 20:01'!
processBackgroundChunk

	| s max val red green blue |
	colorType = 3 ifTrue: [
		backColor _ palette at: chunk first + 1.
		^ self].

	s _ ReadStream on: chunk.
	max _ (2 raisedTo: bitsPerChannel) - 1.

	(colorType = 0 or: [colorType = 4]) ifTrue: [
		val _ s uint16.
		backColor _ Color gray: val / max.
		^ self].

	(colorType = 2 or: [colorType = 6]) ifTrue: [
		red _ s uint16.
		green _ s uint16.
		blue _ s uint16.
		backColor _ Color r: red / max g: green / max b: blue / max.
		^ self].

"====
The bKGD chunk specifies a default background color to present the image against. Note that viewers are not bound to honor this chunk; a viewer can choose to use a different background.

For color type 3 (indexed color), the bKGD chunk contains:

	Palette index: 1 byte

The value is the palette index of the color to be used as background.

For color types 0 and 4 (grayscale, with or without alpha), bKGD contains:

	Gray: 2 bytes, range 0 .. (2^bitdepth)-1

(For consistency, 2 bytes are used regardless of the image bit depth.) The value is the gray level to be used as background.

For color types 2 and 6 (truecolor, with or without alpha), bKGD contains:

	Red: 2 bytes, range 0 .. (2^bitdepth)-1
	Green: 2 bytes, range 0 .. (2^bitdepth)-1
	Blue: 2 bytes, range 0 .. (2^bitdepth)-1

(For consistency, 2 bytes per sample are used regardless of the image bit depth.) This is the RGB color to be used as background.

When present, the bKGD chunk must precede the first IDAT chunk, and must follow the PLTE chunk, if any.
==="
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'RAA 11/4/2000 17:00'!
processIDATChunk

	interlaceMethod = 0
		ifTrue: [ self processNonInterlaced ]
		ifFalse: [ self processInterlaced ]
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'jm 3/17/2009 15:20'!
processIHDRChunk

	| s compression filterMethod |
	s _ ReadStream on: chunk.
	width _ s uint32.
	height _ s uint32.
	bitsPerChannel _ s next.
	colorType _ s next.
	compression _ s next.		"TODO - validate compression"
	filterMethod _ s next.		"TODO - validate filterMethod"
	interlaceMethod _ s next.	"TODO - validate interlace method"

	(#(2 4 6) includes: colorType) ifTrue: [depth _ 32].

	(#(0 3) includes: colorType) ifTrue: [
		depth _ bitsPerChannel min: 8.
		colorType = 0 ifTrue: [ "grayscale"
			palette _ self grayColorsFor: depth]].

	bitsPerPixel _ (BPP at: colorType + 1) at: bitsPerChannel highBit.
	bytesPerScanline _ ((width * bitsPerPixel) + 7) // 8.
	rowSize _ ((width * depth) + 31) >> 5.
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'jm 12/23/2008 18:34'!
processInterlaced

	| z filter bytesPerPass startingCol colIncrement rowIncrement startingRow cx sc temp |
	startingCol _ #(0 4 0 2 0 1 0).
	colIncrement _ #(8 8 4 4 2 2 1).
	rowIncrement _ #(8 8 8 4 4 2 2).
	startingRow _ #(0 0 4 0 2 0 1).

	z _ ZLibReadStream on: chunk from: 1 to: chunk size.
	1 to: 7 do: [:pass |
		(self doPass: pass) ifTrue: [
			cx _ colIncrement at: pass.
			sc _ startingCol at: pass.
			bytesPerPass _ width - sc + cx - 1 // cx * bitsPerPixel + 7 // 8.
			prevScanline _ ByteArray new: bytesPerPass.
			thisScanline _ ByteArray new: bytesPerScanline.
			(startingRow at: pass) to: height - 1 by: (rowIncrement at: pass) do: [:y |
				filter _ z next.
				filtersSeen add: filter.
				(filter isNil or: [(filter between: 0 and: 4) not])
					ifTrue: [^ self].
				thisScanline _ z next: bytesPerPass into: thisScanline startingAt: 1.
				self filterScanline: filter count: bytesPerPass.
				self copyPixels: y at: sc by: cx.
				temp _ prevScanline.
				prevScanline _ thisScanline.
				thisScanline _ temp]]].

	z atEnd ifFalse: [self error: 'Unexpected data'].
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'jm 3/17/2009 15:15'!
processNextChunk

	| length chunkType chunkCrc crc |
	length _ stream uint32.
	chunkType _ (self next: 4) asString.
	chunk _ self next: length.

	chunkCrc _ stream uint32 bitXor: 16rFFFFFFFF.
	crc _ self updateCrc: 16rFFFFFFFF from: 1 to: 4 in: chunkType.
	crc _ self updateCrc: crc from: 1 to: length in: chunk.
	crc = chunkCrc ifFalse: [
		self error: 'PNGReadWriter crc error in chunk ', chunkType].

	chunkType = 'IEND' ifTrue: [^ self].		"*should* be the last chunk"
	chunkType = 'sBIT' ifTrue: [^ self processSBITChunk].	"could indicate unusual sample depth in original"
	chunkType = 'gAMA' ifTrue: [^ self].	"indicates gamma correction value"
	chunkType = 'bKGD' ifTrue: [^ self processBackgroundChunk].
	chunkType = 'pHYs' ifTrue: [^ self processPhysicalPixelChunk].
	chunkType = 'tRNS' ifTrue: [^ self processTransparencyChunk].
	chunkType = 'IHDR' ifTrue: [^ self processIHDRChunk].
	chunkType = 'PLTE' ifTrue: [^ self processPLTEChunk].
	chunkType = 'IDAT' ifTrue: [
		"Since the compressed data can span multiple chunks, stitch them all together first. Later, if memory is an issue, we need to figure out how to do this on the fly."
		globalDataChunk _ globalDataChunk
			ifNil: [chunk]
			ifNotNil: [globalDataChunk,chunk].
		^ self].
	unknownChunks add: chunkType.
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'jm 3/16/2009 20:10'!
processNonInterlaced

	| z filter temp copyMethod debug |
	debug _ self debugging.
	copyMethod _ #(
		copyPixelsGray: nil copyPixelsRGB: copyPixelsIndexed:
		copyPixelsGrayAlpha: nil copyPixelsRGBA:) at: colorType + 1.

	debug ifTrue: [Transcript cr; nextPutAll: 'NI chunk size='; print: chunk size].

	z _ ZLibReadStream on: chunk from: 1 to: chunk size.
	prevScanline _ ByteArray new: bytesPerScanline.
	thisScanline _ ByteArray new: bytesPerScanline.
	0 to: height-1 do: [:y |
		filter _ (z next: 1) first.
		debug ifTrue: [filtersSeen add: filter].
		thisScanline _ z next: bytesPerScanline into: thisScanline startingAt: 1.

		(debug and: [thisScanline size < bytesPerScanline]) ifTrue: [
			Transcript nextPutAll:
				'wanted ', bytesPerScanline printString,
				' but only got ', thisScanline size printString,
				' bytes for this scanline'; cr].

		filter = 0 ifFalse: [self filterScanline: filter count: bytesPerScanline].
		self perform: copyMethod with: y.
		temp _ prevScanline.
		prevScanline _ thisScanline.
		thisScanline _ temp].

	z atEnd ifFalse: [self error:'Unexpected data'].

	debug ifTrue: [Transcript nextPutAll: ' compressed size='; print: z position].
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'jm 12/23/2008 18:29'!
processPLTEChunk

	| colorCount i |
	colorCount _ chunk size // 3.  "TODO - validate colorCount against depth"
	palette _ Array new: colorCount.
	0 to: colorCount - 1 do: [:index |
		i _ (index * 3) + 1.
		palette at: index + 1 put:
			(Color
				r: (chunk at: i) / 255.0
				g: (chunk at: i + 1) / 255.0
				b: (chunk at: i + 2) / 255.0)].
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'RAA 11/5/2000 11:24'!
processPhysicalPixelChunk

	"Transcript show: '  PHYSICAL: ',chunk printString."
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'ar 12/12/2003 18:33'!
processSBITChunk

	| rBits gBits bBits aBits |
	colorType = 6 ifFalse: [^ self].
	rBits _ chunk at: 1.
	gBits _ chunk at: 2.
	bBits _ chunk at: 3.
	aBits _ chunk at: 4.
	(rBits = 5 and: [gBits = 5 and: [bBits = 5 and: [aBits = 1]]])
		ifTrue: [depth _ 16].
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'jm 12/9/2003 18:39'!
processTransparencyChunk

	| s red green blue |
	s _ ReadStream on: chunk.

	colorType = 0 ifTrue: [
		transparentPixelValue _ s uint16.
		^ self].

	colorType = 2 ifTrue: [
		red _ chunk at: 2.
		green _ chunk at: 2.
		blue _ chunk at: 2.
		transparentPixelValue _ 16rFF00 + red << 8 + green << 8 + blue.
		^ self].

	colorType = 3 ifTrue: [
		chunk withIndexDo: [:alpha :index |
			palette at: index put: ((palette at: index) alpha: alpha / 255)].
		^ self].
! !


!PNGReadWriter methodsFor: 'filtering' stamp: 'RAA 11/7/2000 09:43'!
filterAverage: count
	"Use the average of the pixel to the left and the pixel above as a predictor"

	| delta |
	delta _ bitsPerPixel // 8 max: 1.
	1 to: delta do: [:i |
		thisScanline at: i put: ((thisScanline at: i) + ((prevScanline at: i) // 2) bitAnd: 255)].
	delta + 1 to: count do: [:i |
		thisScanline at: i put:
			((thisScanline at: i)
			+ ((prevScanline at: i)
			+ (thisScanline at: i - delta) // 2) bitAnd: 255)]! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'DSM 3/25/2000 17:54'!
filterHorizontal: count
	"Use the pixel to the left as a predictor"

	| delta |
	delta _ bitsPerPixel // 8 max: 1.
	delta+1 to: count do: [ :i |
		thisScanline at: i put: (((thisScanline at: i) +
(thisScanline at: i-delta)) bitAnd: 255) ]


! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'DSM 3/25/2000 17:55'!
filterNone: count
! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'RAA 11/7/2000 09:45'!
filterPaeth: count
	"Select one of (the pixel to the left, the pixel above and the pixel to above left) to
	predict the value of this pixel"

	| delta |
	delta _ bitsPerPixel // 8 max: 1.
	1 to: delta do: [ :i |
		thisScanline at: i put:
			(((thisScanline at: i) + (prevScanline at: i)) bitAnd: 255)].
	delta+1 to: count do: [ :i |
		thisScanline
			at: i
			put: (((thisScanline at: i) + (self
				paethPredictLeft: (thisScanline at: i-delta)
				above: (prevScanline at: i)
				aboveLeft: (prevScanline at: i-delta)))
					bitAnd: 255)]

! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'eat 9/11/2000 20:08'!
filterScanline: filterType count: count

	self
		perform: (
			#(filterNone: filterHorizontal: filterVertical: filterAverage: filterPaeth:)
				at: filterType+1)
		with: count.

! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'DSM 3/25/2000 17:54'!
filterVertical: count
	"Use the pixel above as a predictor"

	1 to: count do: [ :i |
		thisScanline at: i put: (((thisScanline at: i) +
(prevScanline at: i)) bitAnd: 255) ]

! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'eat 9/11/2000 20:05'!
paethPredictLeft: a above: b aboveLeft: c
	"Predicts the value of a pixel based on nearby pixels, based on
Paeth (GG II, 1991)"

	| pa pb pc |
	pa _ b > c ifTrue: [b - c] ifFalse: [c - b].
	pb _ a > c ifTrue: [a - c] ifFalse: [c - a].
	pc _ a + b - c - c.
	pc < 0 ifTrue: [
		pc := pc * -1].
	((pa <= pb) and: [pa <= pc]) ifTrue: [^ a].
	(pb <= pc) ifTrue: [^ b].
	^ c
! !


!PNGReadWriter methodsFor: 'pixel copies' stamp: 'DSM 3/26/2000 21:32'!
copyPixels: y
	"Handle non-interlaced pixels of supported colorTypes"

	| s |
	s _ #(copyPixelsGray: nil copyPixelsRGB: copyPixelsIndexed:
		  copyPixelsGrayAlpha: nil copyPixelsRGBA:) at: colorType+1.
	self perform: s asSymbol with: y
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/4/2000 16:08'!
copyPixels: y at: startX by: incX
	"Handle interlaced pixels of supported colorTypes"

	| s |
	s _ #(copyPixelsGray:at:by: nil copyPixelsRGB:at:by: copyPixelsIndexed:at:by:
		  copyPixelsGrayAlpha:at:by: nil copyPixelsRGBA:at:by:) at: colorType+1.
	self perform: s asSymbol with: y with: startX with: incX
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'jm 6/21/2003 11:23'!
copyPixelsGray: y 
	"Handle non-interlaced grayscale color mode (colorType = 0)"
	| blitter pixPerByte mask shifts pixelNumber rawByte pixel transparentIndex |
	blitter _ BitBlt bitPokerToForm: form.
	transparentIndex _ form colors size.
	bitsPerChannel = 16
		ifTrue: [0
				to: width - 1
				do: [:x | blitter pixelAt: x @ y put: 255
							- (thisScanline at: x << 1 + 1)].
			^ self]
		ifFalse: [bitsPerChannel = 8
				ifTrue: [1
						to: width
						do: [:x | blitter
								pixelAt: x - 1 @ y
								put: (thisScanline at: x)].
					^ self].
			bitsPerChannel = 1
				ifTrue: [pixPerByte _ 8.
					mask _ 1.
					shifts _ #(7 6 5 4 3 2 1 0 )].
			bitsPerChannel = 2
				ifTrue: [pixPerByte _ 4.
					mask _ 3.
					shifts _ #(6 4 2 0 )].
			bitsPerChannel = 4
				ifTrue: [pixPerByte _ 2.
					mask _ 15.
					shifts _ #(4 0 )].
			pixelNumber _ 0.
			0 to: width - 1 do: [:x | 
				rawByte _ thisScanline at: pixelNumber // pixPerByte + 1.
				pixel _ rawByte
							>> (shifts at: pixelNumber \\ pixPerByte + 1) bitAnd: mask.
				pixel = transparentPixelValue ifTrue: [pixel _ transparentIndex].
				blitter pixelAt: x @ y put: pixel.
				pixelNumber _ pixelNumber + 1
			]
		]! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'jm 6/21/2003 11:24'!
copyPixelsGray: y at: startX by: incX
	"Handle interlaced grayscale color mode (colorType = 0)"

	| b offset bits w pixel mask blitter pixelNumber pixPerByte rawByte
shifts |
	bitsPerChannel = 16
		ifTrue: [
			b _ BitBlt bitPokerToForm: form.
			startX to: width-1 by: incX do: [ :x |
				b pixelAt: x@y put: 255 - (thisScanline at: (x//incX<<1)+1).
				].
			^ self
			].
	offset _ y*rowSize+1.
	bits _ form bits.
	bitsPerChannel = 8 ifTrue: [
		startX to: width-1 by: incX do: [ :x |
			w _ offset + (x>>2).
			b _ 3- (x \\ 4) * 8.
			pixel _ (thisScanline at: x // incX + 1)<<b.
			mask _ (255<<b) bitInvert32.
			bits at: w put: (((bits at: w) bitAnd: mask) bitOr: pixel)
		].
		^ self
	].
	bitsPerChannel = 1 ifTrue: [
		pixPerByte _ 8.
		mask _ 1.
		shifts _ #(7 6 5 4 3 2 1 0).
	].
	bitsPerChannel = 2 ifTrue: [
		pixPerByte _ 4.
		mask _ 3.
		shifts _ #(6 4 2 0).
	].
	bitsPerChannel = 4 ifTrue: [
		pixPerByte _ 2.
		mask _ 15.
		shifts _ #(4 0).
	].

	blitter _ BitBlt bitPokerToForm: form.
	pixelNumber _ 0.
	startX to: width-1 by: incX do: [ :x |
		rawByte _ thisScanline at: (pixelNumber // pixPerByte) + 1.
		pixel _ (rawByte >> (shifts at: (pixelNumber \\ pixPerByte) + 1)) bitAnd: mask.
		blitter pixelAt: (x@y) put: pixel.
		pixelNumber _ pixelNumber + 1.
	].
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'jm 9/27/2006 17:13'!
copyPixelsGrayAlpha: y
	"Handle non-interlaced grayscale with alpha color mode (colorType = 4)"

	| b i alpha gray pixel |
	b _ BitBlt bitPokerToForm: form.
	bitsPerChannel = 8
		ifTrue: [
			0 to: width-1 do: [ :x |
				i _ (x << 1) + 1.
				(alpha _ thisScanline at: i+1) > 0
					ifTrue: [
						gray _ thisScanline at: i.
						pixel _ (alpha<<24) + (gray<<16) + (gray<<8) + gray]
					ifFalse: [pixel _ 0].
				b pixelAt: x@y put: pixel.
				]
			]
		ifFalse: [
			0 to: width-1 do: [ :x |
				i _ (x << 2) + 1.
				(alpha _ thisScanline at: i+2) > 0
					ifTrue: [
						gray _ thisScanline at: i.
						pixel _ (alpha<<24) + (gray<<16) + (gray<<8) + gray]
					ifFalse: [pixel _ 0].
				b pixelAt: x@y put: pixel.
				]
			]
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'jm 9/27/2006 17:16'!
copyPixelsGrayAlpha: y at: startX by: incX
	"Handle interlaced grayscale with alpha color mode (colorType = 4)"

	| b i alpha gray pixel |
	b _ BitBlt bitPokerToForm: form.
	bitsPerChannel = 8
		ifTrue: [
			startX to: width-1 by: incX do: [ :x |
				i _ (x // incX << 1) + 1.
				(alpha _ thisScanline at: i+1) > 0
					ifTrue: [
						gray _ thisScanline at: i.
						pixel _ (alpha<<24) + (gray<<16) + (gray<<8) + gray]
					ifFalse: [pixel _ 0].
				b pixelAt: x@y put: pixel.
				]
			]
		ifFalse: [
			startX to: width-1 by: incX do: [ :x |
				i _ (x // incX << 2) + 1.
				(alpha _ thisScanline at: i+2) > 0
					ifTrue: [
						gray _ thisScanline at: i.
						pixel _ (alpha<<24) + (gray<<16) + (gray<<8) + gray]
					ifFalse: [pixel _ 0].
				b pixelAt: x@y put: pixel.
				]
			]
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'jm 6/21/2003 11:23'!
copyPixelsIndexed: y
	"Handle non-interlaced indexed color mode (colorType = 3)"
	| blitter pixPerByte mask shifts pixelNumber rawByte pixel |
	blitter _ BitBlt bitPokerToForm: form.
	bitsPerChannel = 8
		ifTrue:
			[1 to: width do: [:x | blitter pixelAt: x - 1 @ y put: (thisScanline at: x)].
			^ self].
	bitsPerChannel = 1
		ifTrue:
			[pixPerByte _ 8.
			mask _ 1.
			shifts _ #(7 6 5 4 3 2 1 0 )].
	bitsPerChannel = 2
		ifTrue:
			[pixPerByte _ 4.
			mask _ 3.
			shifts _ #(6 4 2 0 )].
	bitsPerChannel = 4
		ifTrue:
			[pixPerByte _ 2.
			mask _ 15.
			shifts _ #(4 0 )].
	pixelNumber _ 0.
	0 to: width - 1 do:
		[:x |
		rawByte _ thisScanline at: pixelNumber // pixPerByte + 1.
		pixel _ rawByte >> (shifts at: pixelNumber \\ pixPerByte + 1) bitAnd: mask.
		blitter pixelAt: x @ y put: pixel.
		pixelNumber _ pixelNumber + 1]! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'jm 6/21/2003 11:23'!
copyPixelsIndexed: y at: startX by: incX
	"Handle interlaced indexed color mode (colorType = 3)"

	| offset b bits w pixel mask pixPerByte shifts blitter pixelNumber rawByte |
	offset _ y*rowSize+1.
	bits _ form bits.
	bitsPerChannel = 8
		ifTrue: [
			startX to: width-1 by: incX do: [ :x |
				w _ offset + (x>>2).
				b _ 3 - (x \\ 4) * 8.
				pixel _ (thisScanline at: x // incX + 1)<<b.
				mask _ (255<<b) bitInvert32.
				bits at: w put: (((bits at: w) bitAnd: mask) bitOr: pixel)].
			^ self ].
	bitsPerChannel = 1 ifTrue: [
		pixPerByte _ 8.
		mask _ 1.
		shifts _ #(7 6 5 4 3 2 1 0).
	].
	bitsPerChannel = 2 ifTrue: [
		pixPerByte _ 4.
		mask _ 3.
		shifts _ #(6 4 2 0).
	].
	bitsPerChannel = 4 ifTrue: [
		pixPerByte _ 2.
		mask _ 15.
		shifts _ #(4 0).
	].

	blitter _ BitBlt bitPokerToForm: form.
	pixelNumber _ 0.
	startX to: width-1 by: incX do: [ :x |
		rawByte _ thisScanline at: (pixelNumber // pixPerByte) + 1.
		pixel _ (rawByte >> (shifts at: (pixelNumber \\ pixPerByte) + 1)) bitAnd: mask.
		blitter pixelAt: (x@y) put: pixel.
		pixelNumber _ pixelNumber + 1.
	].
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/7/2000 09:30'!
copyPixelsRGB: y
	"Handle non-interlaced RGB color mode (colorType = 2)"

	| i pixel tempForm tempBits |

	tempForm _ Form extent: width@1 depth: 32.
	tempBits _ tempForm bits.
	pixel := LargePositiveInteger new: 4.
	pixel at: 4 put: 16rFF.
	bitsPerChannel = 8 ifTrue: [
		i := 1.
		1 to: width do: [ :x |
			pixel
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+1);
				at: 1 put: (thisScanline at: i+2).
			tempBits at: x put: pixel.
			i _ i + 3.
		]
	] ifFalse: [
		i := 1.
		1 to: width do: [ :x |
			pixel
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+2);
				at: 1 put: (thisScanline at: i+4).
			tempBits at: x put: pixel.
			i _ i + 6.
		]
	].
	transparentPixelValue ifNotNil: [
		1 to: width do: [ :x |
			(tempBits at: x) = transparentPixelValue ifTrue: [
				tempBits at: x put: 0.
			].
		].
	].
	tempForm displayOn: form at: 0@y rule: Form paint.
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/5/2000 11:16'!
copyPixelsRGB: y at: startX by: incX
	"Handle interlaced RGB color mode (colorType = 2)"

	| i pixel tempForm tempBits xx loopsToDo |

	tempForm _ Form extent: width@1 depth: 32.
	tempBits _ tempForm bits.
	pixel := LargePositiveInteger new: 4.
	pixel at: 4 put: 16rFF.
	loopsToDo _ width - startX + incX - 1 // incX.
	bitsPerChannel = 8 ifTrue: [
		i _ (startX // incX * 3) + 1.
		xx _ startX+1.
		1 to: loopsToDo do: [ :j |
			pixel
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+1);
				at: 1 put: (thisScanline at: i+2).
			tempBits at: xx put: pixel.
			i _ i + 3.
			xx _ xx + incX.
		]
	] ifFalse: [
		i _ (startX // incX * 6) + 1.
		xx _ startX+1.
		1 to: loopsToDo do: [ :j |
			pixel
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+2);
				at: 1 put: (thisScanline at: i+4).
			tempBits at: xx put: pixel.
			i _ i + 6.
			xx _ xx + incX.
		].
	].
	transparentPixelValue ifNotNil: [
		startX to: width-1 by: incX do: [ :x |
			(tempBits at: x+1) = transparentPixelValue ifTrue: [
				tempBits at: x+1 put: 0.
			].
		].
	].
	tempForm displayOn: form at: 0@y rule: Form paint.

! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'jm 9/27/2006 17:22'!
copyPixelsRGBA: y
	"Handle non-interlaced RGBA color modes (colorType = 6)"

	| tempForm tempBits pixel i alpha |
	tempForm _ Form extent: width@1 depth: 32.
	tempBits _ tempForm bits.
	pixel := LargePositiveInteger new: 4.
	bitsPerChannel = 8 ifTrue: [
		i := -3.
		0 to: width-1 do: [ :x |
			i := i + 4.
			(alpha _ thisScanline at: i+3) > 0
				ifTrue: [
					pixel
						at: 4 put: alpha;
						at: 3 put: (thisScanline at: i);
						at: 2 put: (thisScanline at: i+1);
						at: 1 put: (thisScanline at: i+2).
					tempBits at: x+1 put: pixel]
				ifFalse: [
					tempBits at: x+1 put: 0].
		]
	] ifFalse: [
		i := -7.
		0 to: width-1 do: [ :x |
			i := i + 8.
			(alpha _ thisScanline at: i+6) > 0
				ifTrue: [
					pixel
						at: 4 put: alpha;
						at: 3 put: (thisScanline at: i);
						at: 2 put: (thisScanline at: i+2);
						at: 1 put: (thisScanline at: i+4).
					tempBits at: x+1 put: pixel]
				ifFalse: [
					tempBits at: x+1 put: 0].
		]
	].
	tempForm displayOn: form at: 0@y rule: Form paint.
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'jm 9/27/2006 17:27'!
copyPixelsRGBA: y at: startX by: incX
	"Handle interlaced RGBA color modes (colorType = 6)"

	| tempForm tempBits pixel i alpha |
	tempForm _ Form extent: width@1 depth: 32.
	tempBits _ tempForm bits.
	pixel := LargePositiveInteger new: 4.
	bitsPerChannel = 8 ifTrue: [
		i _ (startX // incX << 2) + 1.
		startX to: width-1 by: incX do: [ :x |
			(alpha _ thisScanline at: i+3) > 0
				ifTrue: [
					pixel
						at: 4 put: alpha;
						at: 3 put: (thisScanline at: i);
						at: 2 put: (thisScanline at: i+1);
						at: 1 put: (thisScanline at: i+2).
					tempBits at: x+1 put: pixel]
				ifFalse: [
					tempBits at: x+1 put: 0].
			i _ i + 4.
		]
	] ifFalse: [
		i _ (startX // incX << 3) +1.
		startX to: width-1 by: incX do: [ :x |
			(alpha _ thisScanline at: i+6) > 0
				ifTrue: [
					pixel
						at: 4 put: alpha;
						at: 3 put: (thisScanline at: i);
						at: 2 put: (thisScanline at: i+2);
						at: 1 put: (thisScanline at: i+4).
					tempBits at: x+1 put: pixel]
				ifFalse: [
					tempBits at: x+1 put: 0].
			i _ i + 8.
		].
	].
	tempForm displayOn: form at: 0@y rule: Form paint.

! !


!PNGReadWriter methodsFor: 'miscellaneous' stamp: 'DSM 4/27/2000 13:09'!
doPass: pass
	"Certain interlace passes are skipped with certain small image
dimensions"

	pass = 1 ifTrue: [ ^ true ].
	((width = 1) and: [height = 1]) ifTrue: [ ^ false ].
	pass = 2 ifTrue: [ ^ width >= 5 ].
	pass = 3 ifTrue: [ ^ height >= 5 ].
	pass = 4 ifTrue: [ ^ (width >=3 ) or: [height >= 5] ].
	pass = 5 ifTrue: [ ^ height >=3 ].
	pass = 6 ifTrue: [ ^ width >=2 ].
	pass = 7 ifTrue: [ ^ height >=2 ].

! !

!PNGReadWriter methodsFor: 'miscellaneous' stamp: 'jm 12/23/2008 18:24'!
grayColorsFor: d
	"return a color table for a gray image"

	palette _ Array new: 1 << d.
	d = 1 ifTrue: [
		palette at: 1 put: Color black.
		palette at: 2 put: Color white.
		^ palette copyWith: Color transparent].

	d = 2 ifTrue: [
		palette at: 1 put: Color black.
		palette at: 2 put: (Color gray: 85.0 / 255.0).
		palette at: 3 put: (Color gray: 170.0 / 255.0).
		palette at: 4 put: Color white.
		^ palette copyWith: Color transparent].

	d = 4 ifTrue: [
		0 to: 15 do: [:g |
			palette at: g + 1 put: (Color gray: g / 15.0)].
		^ palette copyWith: Color transparent].

	d = 8 ifTrue: [
		0 to: 255 do: [:g |
			palette at: g + 1 put: (Color gray: g / 255.0)].
		^ palette].
! !


!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 16:37'!
nextPutImage: aForm
	"Write out the given form. We're keeping it simple here, no interlacing, no filters."

	^ self nextPutImage: aForm interlace: 0 filter: 0  "no filtering"
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'jens 11/21/2008 12:42'!
nextPutImage: aForm interlace: aMethod filter: aFilterType 
	"Note: For now we keep it simple - interlace and filtering are simply ignored."

	| crcStream |
	bigEndian _ Smalltalk isBigEndian.
	form _ aForm.
	width _ aForm width.
	height _ aForm height.
	aForm depth <= 8 
		ifTrue: 
			[bitsPerChannel _ aForm depth.
			colorType _ 3.
			bytesPerScanline _ (width * aForm depth + 7) // 8]
		ifFalse: 
			[bitsPerChannel _ 8.
			colorType _ 6.
			bytesPerScanline _ width * 4].

	self writeFileSignature.
	crcStream _ WriteStream on: (ByteArray new: 1000).
	crcStream resetToStart.
	self writeIHDRChunkOn: crcStream.
	self writeChunk: crcStream.

	form depth <= 8 ifTrue: [
		crcStream resetToStart.
		self writePLTEChunkOn: crcStream.
		self writeChunk: crcStream.
		(form isKindOf: ColorForm) ifTrue: [
			crcStream resetToStart.
			self writeTRNSChunkOn: crcStream.
			self writeChunk: crcStream]].

	form depth = 16 ifTrue: [
		crcStream resetToStart.
		self writeSBITChunkOn: crcStream.
		self writeChunk: crcStream].

	crcStream resetToStart.
	self writeIDATChunkOn: crcStream.
	self writeChunk: crcStream.
	crcStream resetToStart.
	self writeIENDChunkOn: crcStream.
	self writeChunk: crcStream
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'jens 11/18/2008 21:04'!
updateCrc: oldCrc from: start to: stop in: aCollection

	^ (GZipWriteStream on: String new)
		updateCrc: oldCrc from: start to: stop in: aCollection
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'jm 12/23/2008 15:56'!
writeChunk: crcStream

	| bytes length crc debug |
	debug _ self debugging.
	bytes _ crcStream originalContents.
	length _ crcStream position.
	crc _ self updateCrc: 16rFFFFFFFF from: 1 to: length in: bytes.
	crc _ crc bitXor: 16rFFFFFFFF.

	debug ifTrue: [
		Transcript
			cr;
			print: stream position; space;
			nextPutAll: (bytes copyFrom: 1 to: 4) asString;
			nextPutAll: ' len='; print: length;
			nextPutAll: ' crc=0x'; nextPutAll: crc hex].

	stream uint32: length-4. "exclude chunk name"
	stream next: length putAll: bytes startingAt: 1.
	stream uint32: crc.
	debug ifTrue: [Transcript nextPutAll: ' afterPos='; print: stream position].
	crcStream resetToStart.
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 16:40'!
writeFileSignature

	stream nextPutAll: #(16r89 16r50 16r4E  16r47 16r0D 16r0A 16r1A 16r0A) asByteArray.
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'nk 2/17/2004 14:57'!
writeIDATChunkOn: aStream
	"Write the IDAT chunk."

	| z |
	aStream nextPutAll: 'IDAT' asByteArray.
	z _ ZLibWriteStream on: aStream.
	form depth <= 8 
		ifTrue: [self writeType3DataOn: z]
		ifFalse: [self writeType6DataOn: z].

	self debugging ifTrue: [
		Transcript cr;
			nextPutAll: 'compressed size=';
			print: aStream position;
			nextPutAll: ' uncompressed size=';
			print: z position].
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 17:08'!
writeIENDChunkOn: aStream
	"Write the IEND chunk."

	aStream nextPutAll: 'IEND' asByteArray.
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'jm 12/23/2008 15:58'!
writeIHDRChunkOn: aStream
	"Write the IHDR chunk."

	aStream nextPutAll: 'IHDR' asByteArray.
	aStream int32: width.
	aStream int32: height.
	aStream nextPut: bitsPerChannel.
	aStream nextPut: colorType.
	aStream nextPut: 0. "compression"
	aStream nextPut: 0. "filter method"
	aStream nextPut: 0. "interlace method"
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'jens 11/21/2008 12:42'!
writePLTEChunkOn: aStream
	"Write the PLTE chunk."

	| r g b colors |
	aStream nextPutAll: 'PLTE' asByteArray.
	(form isKindOf: ColorForm) 
		ifTrue: [colors _ form colors]
		ifFalse: [colors _ Color indexedColors copyFrom: 1 to: (1 bitShift: form depth)].
	colors do: [:aColor |
		r _ (aColor red * 255) truncated.
		g _ (aColor green * 255) truncated.
		b _ (aColor blue * 255) truncated.
		aStream nextPut: r; nextPut: g; nextPut: b.
	].
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 18:29'!
writeSBITChunkOn: aStream
	"Write the IDAT chunk."

	aStream nextPutAll: 'sBIT' asByteArray.
	form depth = 16 ifFalse: [self error: 'Unimplemented feature'].
	aStream nextPut: 5.
	aStream nextPut: 5.
	aStream nextPut: 5.
	aStream nextPut: 1.
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 17:34'!
writeTRNSChunkOn: aStream
	"Write out tRNS chunk."

	aStream nextPutAll: 'tRNS' asByteArray.
	form colors do: [:aColor|
		aStream nextPut: (aColor alpha * 255) truncated.
	].
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 1/1/1970 20:58'!
writeType3DataOn: zStream
	"Write color indexed data."

	| scanline hack hackBlt swizzleBlt swizzleHack hackDepth |
	scanline _ ByteArray new: bytesPerScanline + 3 // 4 * 4.
	hackDepth _ bigEndian ifTrue: [form depth] ifFalse: [form depth negated].
	hack _ Form extent: width@1 depth: hackDepth bits: scanline.
	hackBlt _ BitBlt toForm: hack.
	hackBlt sourceForm: form.
	hackBlt combinationRule: Form over.
	hackBlt destOrigin: 0@0.
	hackBlt width: width; height: 1.
	(form depth < 8 and: [bigEndian not]) ifTrue: [
		swizzleHack _ Form new hackBits: scanline.
		swizzleBlt _ BitBlt toForm: swizzleHack.
		swizzleBlt sourceForm: swizzleHack.
		swizzleBlt combinationRule: Form over.
		swizzleBlt colorMap: (StandardSwizzleMaps at: form depth)].

	0 to: height-1 do: [:i |
		hackBlt sourceOrigin: 0@i; copyBits.
		swizzleBlt ifNotNil: [swizzleBlt copyBits].
		zStream nextPut: 0. "filterType"
		zStream next: bytesPerScanline putAll: scanline startingAt: 1].

	zStream close.
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'jm 3/18/2009 07:08'!
writeType6DataOn: zStream
	"Write RGBA data."
	"NOTE: This is a work-around for a bug that reversed R and B. But this code is slower than the BitBlt code it replaces so this should be revisited."

	| transparentRGBA bits pix |
	transparentRGBA _ #(0 0 0 0) asByteArray.
	bits _ form bits.
	form depth < 32 ifTrue: [bits _ (form asFormOfDepth: 32) bits].
	1 to: bits size do: [:i |
		(i \\ form width) = 1 ifTrue: [zStream nextPut: 0]. "insert filterType at start of each scanline"
		pix _ bits at: i.
		pix = 0
			ifTrue: [zStream nextPutAll: transparentRGBA]
			ifFalse: [
				zStream nextPut: ((pix bitAnd: 16rFF0000) bitShift: -16).
				zStream nextPut: ((pix bitAnd: 16rFF00) bitShift: -8).
				zStream nextPut: (pix bitAnd: 16rFF).
				zStream nextPut: 16rFF]].

	zStream close.
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'jm 3/24/2009 16:40'!
writeType6DataOnOLD: zStream
	"Write RGBA data."

	| scanline hack hackBlt cm miscBlt |
	scanline _ ByteArray new: bytesPerScanline.
	hack _ Form extent: width@1 depth: 32 bits: scanline.
	form depth = 16 ifTrue: [
		"Expand 16 -> 32"
		miscBlt _ BitBlt toForm: hack.
		miscBlt sourceForm: form.
		miscBlt combinationRule: Form over.
		miscBlt destOrigin: 0@0.
		miscBlt width: width; height: 1].

	hackBlt _ BitBlt toForm: hack.
	hackBlt sourceForm: (miscBlt ifNil: [form] ifNotNil: [hack]).
	hackBlt combinationRule: Form over.
	hackBlt destOrigin: 0@0.
	hackBlt width: width; height: 1.
	bigEndian
		ifTrue: [
			cm _ ColorMap 
				shifts: #(8 8 8 -24) 
				masks: #(16rFF0000 16rFF00 16rFF 16rFF000000)]
		ifFalse: [
			cm _ ColorMap 
				shifts: #(-16 0 16 0) 
				masks: #(16rFF0000 16rFF00 16rFF 16rFF000000)].

	hackBlt colorMap: cm.
	0 to: height-1 do: [:i |
		miscBlt
			ifNil: [
				hackBlt sourceOrigin: 0@i; copyBits]
			ifNotNil: [
				miscBlt sourceOrigin: 0@i; copyBits.
				hack fixAlpha.
				hackBlt copyBits].

		zStream nextPut: 0. "filterType"
		zStream nextPutAll: scanline.
	].
	zStream close.
! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PNGReadWriter class
	instanceVariableNames: ''!

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'ar 2/11/2004 00:54'!
computeSwizzleMapForDepth: depth
	"Answer a map that maps pixels in a word to their opposite location. Used for 'middle-endian' forms where the byte-order is different from the bit order (good joke, eh?)."

	| map swizzled |
	map _ Bitmap new: 256.
	depth = 4 ifTrue: [
		0 to: 255 do: [:pix |
			swizzled _ 0.
			swizzled _ swizzled bitOr: (((pix bitShift: 0) bitAnd: 15) bitShift: 4).
			swizzled _ swizzled bitOr: (((pix bitShift: -4) bitAnd: 15) bitShift: 0).
			map at: pix + 1 put: swizzled].

		^ ColorMap colors: map].

	depth = 2 ifTrue: [
		0 to: 255 do: [:pix |
			swizzled _ 0.
			swizzled _ swizzled bitOr: (((pix bitShift: 0) bitAnd: 3) bitShift: 6).
			swizzled _ swizzled bitOr: (((pix bitShift: -2) bitAnd: 3) bitShift: 4).
			swizzled _ swizzled bitOr: (((pix bitShift: -4) bitAnd: 3) bitShift: 2).
			swizzled _ swizzled bitOr: (((pix bitShift: -6) bitAnd: 3) bitShift: 0).
			map at: pix + 1 put: swizzled].
		^ ColorMap colors: map].

	depth = 1 ifTrue: [
		0 to: 255 do: [:pix |
			swizzled _ 0.
			swizzled _ swizzled bitOr: (((pix bitShift: 0) bitAnd: 1) bitShift: 7).
			swizzled _ swizzled bitOr: (((pix bitShift: -1) bitAnd: 1) bitShift: 6).
			swizzled _ swizzled bitOr: (((pix bitShift: -2) bitAnd: 1) bitShift: 5).
			swizzled _ swizzled bitOr: (((pix bitShift: -3) bitAnd: 1) bitShift: 4).
			swizzled _ swizzled bitOr: (((pix bitShift: -4) bitAnd: 1) bitShift: 3).
			swizzled _ swizzled bitOr: (((pix bitShift: -5) bitAnd: 1) bitShift: 2).
			swizzled _ swizzled bitOr: (((pix bitShift: -6) bitAnd: 1) bitShift: 1).
			swizzled _ swizzled bitOr: (((pix bitShift: -7) bitAnd: 1) bitShift: 0).
			map at: pix + 1 put: swizzled].
		^ ColorMap colors: map].

	self error: 'Unrecognized depth'
! !

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'RAA 11/7/2000 09:22'!
createAFormFrom: data

	| error f |

	error _ ''.
	f _ [
		self formFromStream: (RWBinaryOrTextStream with: data)
	] ifError: [ :a :b |
		error _ a printString,'  ',b printString.
		(StringMorph contents: error) color: Color red; imageForm
	].
	^{f. error}! !

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'jm 12/23/2008 15:46'!
initialize
	"PNGReadWriter initialize"

	Debugging _ false.
	BPP _ {	#(1 2 4 8 16).
			#(0 0 0 0 0).
			#(0 0 0 24 48).
			#(1 2 4 8 0).
			#(0 0 0 16 32).
			#(0 0 0 0 0).
			#(0 0 0 32 64).
			#(0 0 0 0 0) }.

	BlockHeight _ #(8 8 4 4 2 2 1).
	BlockWidth _ #(8 4 4 2 2 1 1).

	StandardColors := Color indexedColors collect:[:aColor|
		Color 
			r: (aColor red * 255) truncated / 255
			g: (aColor green * 255) truncated / 255
			b: (aColor blue * 255) truncated / 255.
	].

	StandardSwizzleMaps := Array new: 4.
	#(1 2 4) do:[:i| StandardSwizzleMaps at: i put: (self computeSwizzleMapForDepth: i)].! !

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'jm 3/18/2009 07:04'!
writeTestToFileNamed: fileName
	"Write a tiny 32-bit PNG image to a new file with the given name."
	"self writeTestToFileNamed: 'test.png'"

	| form |
	form _ Form extent: 5@1 depth: 32.
	form colorAt: 0@0 put: Color red.
	form colorAt: 1@0 put: Color green.
	form colorAt: 2@0 put: Color blue.
	form colorAt: 3@0 put: Color black.
	form colorAt: 4@0 put: Color white.
	(form magnifyBy: 20) display.

	self putForm: form onFileNamed: fileName.

"Debugging notes:

The generated file does not display correctly on Mac OS X (tested on 10.5.5) and Firefox says it is a broken image. It does open correctly on Windows. On either Mac and Windows, opening and re-saving the image results in a file that does display on Mac OS.

The problem appears to be how the data chunk ('IDAT' = #(73 68 65 84) asByteArray) is compressed. Here is the payload of this chunk as written by Squeak: vs. several other programs (e.g. Preview on Mac and Paint on Windows): 

rgbSqueak:		120 1 99 248 207 0 68 32 200 240 31 136 25 255 131 0 0
rgbMac:			8 29 99 248 207 192 240 31 12 25 254 3 153 140 255 65 0 0 98 184 10 247
rgbWin:			24 87 99 248 207 192 240 31 12 25 254 3 49 227 127 16 0 0 98 184 10 247

Note that Squeak decompressses all three sequences to the same bytes. For example:

	(ZLibReadStream on: #(120 1 99 248 207 0 68 32 200 240 31 136 25 255 131 0 0) asByteArray) upToEnd

It appears that Squeak is using a non-standard zlib setting. See the Deflate/Inflate section of the PNG spec:

  http://www.libpng.org/pub/png/spec/1.2/PNG-Contents.html
"
! !


PNGReadWriter initialize!
