CommandBlockMorph subclass: #ConfigurationBlockMorph
	instanceVariableNames: 'initialName previousName peripheralClass englishCodeSpec nameless '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Enchanting-Blocks'!
!ConfigurationBlockMorph commentStamp: '<historical>' prior: 0!
This more is used in Enchanting to allow you to configure motors and sensors.  It attached to a  ConfigurationHatMorph and defines what the block is.

For example, if you had a ConfigurationHatMorph named "Sensor Port 1", and attached a ConfigurationBlockMorph called "light sensor", then this means that port 1 has a light sensor attached to it.
!


!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 12/18/2010 03:41'!
concreteType
"Look at the settings in the concrete type selection morph and return the english string for the concrete type"

	| typeMorph |

	typeMorph _ self typeMorph.
	typeMorph ifNotNil: [
		^ typeMorph choice.
	].

	^ nil.! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 12/18/2010 03:43'!
concreteType: aString
"Take the string representing the type and set the appropriate morph with the appropriate name"

	| typeMorph |

	typeMorph _ self typeMorph.
	typeMorph ifNotNil: [
		typeMorph choice: aString
	].! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 11/20/2010 21:14'!
currentName

	| nameMorph |

	nameMorph _ self nameMorph.
	nameMorph ifNotNil: [^ nameMorph evaluate] ifNil: [^ nil].! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 7/5/2011 21:08'!
englishCodeSpec
"Get what the code specification is, in English"

	^ englishCodeSpec.! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 7/5/2011 21:08'!
englishCodeSpec: aSpec
"Set what the code specification is, in English"

	englishCodeSpec _ aSpec.! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 11/20/2010 20:50'!
initialName

	^ initialName ! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 10/26/2012 07:05'!
initialName: aName

	| nameMorph |

	initialName _ aName.
	previousName _ aName.

	nameMorph _ self nameMorph.
	nameMorph ifNotNil: [
		nameMorph stringExpression: initialName.

		"Ugly.  Ask the StringFieldMorph to notify us of changes."
		nameMorph submorphs first client: self; acceptWhenFocusLost: true.
	].
! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 11/13/2010 15:43'!
isStop
"While this is not a block that says 'execution stops here', it does stop a script -- the whole script is always a hat block and a single configuration block in the configuration menu."

	^ true! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 10/26/2012 07:04'!
nameMorph
	"Returns the morph in which the user can edit the name (or nil)"

	nameless ifTrue: [ ^ nil ].

	^ submorphs detect: [
		:s | (s isKindOf: ExpressionArgMorph)
	] ifNone: [ nil ].
	! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 10/26/2012 07:03'!
nameless
"Is this a block that isn't supposed to have a name?"

	^ nameless.! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 10/26/2012 07:03'!
nameless: isNameless
"Is this a block that isn't supposed to have a name?"

	nameless _ isNameless.! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 11/20/2010 20:50'!
peripheralClass

	^ peripheralClass ! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 11/20/2010 20:51'!
peripheralClass: aClass

	peripheralClass _ aClass! !

!ConfigurationBlockMorph methodsFor: 'accessing' stamp: 'CWB 12/16/2010 16:31'!
typeMorph
	"Returns the morph in which the user chooses the concrete subtype"

	^ submorphs detect: [:m | 
		(m isKindOf: ChoiceArgMorphForPeripheral) 
		and: [ m getOptionsSelector = #concreteSubtypesFor: ] ] 
	ifNone: [ nil ].
	! !


!ConfigurationBlockMorph methodsFor: 'converting' stamp: 'CWB 11/22/2010 20:31'!
createPeripheral
"Creates a peripheral that matches the representation this block shows."

	^ peripheralClass new takeSettingsFrom: self.
	! !


!ConfigurationBlockMorph methodsFor: 'dropping/grabbing' stamp: 'CWB 12/11/2010 10:46'!
justDroppedInto: newOwner event: evt
	"Handle being dropped into a new situation."

	| frame |

	(frame _ newOwner ownerThatIsA: ScratchFrameMorph)
		ifNotNil: [frame projectModified].

	(newOwner ownerThatIsA: EnchantingConfigurationDialog) ifNotNil:
	[
		"If we are dropped into the block bin on the dialog, stay"
		(newOwner isKindOf: ScratchScriptsMorph) ifTrue: [ ^self handleDropOn: newOwner ].

		"If we are dropped onto the palette area on the dialog, disappear"
		(newOwner isKindOf: ScratchBlockPaletteMorph) ifTrue: [ self delete. ^self ].
	].

	"If we are dropped anywhere else, reject the drop"
	^ self rejectDropEvent: evt.
! !


!ConfigurationBlockMorph methodsFor: 'initialization' stamp: 'CWB 7/5/2011 21:08'!
initialize

	super initialize.
	peripheralClass _ nil.
	initialName _ nil.
	previousName _ nil.
	englishCodeSpec _ nil.
	! !


!ConfigurationBlockMorph methodsFor: 'private' stamp: 'CWB 12/31/2012 11:31'!
wrapWidth
"We want to wrap blocks that are wider than this"
	
	^ 175.! !


!ConfigurationBlockMorph methodsFor: 'event handling' stamp: 'CWB 1/1/2011 12:28'!
acceptContents
"Callback from the name label morph when things have been edited"

	"Unregister the previous name.  Get a new name, show it, and register it."

	| dlg newName |

	(dlg _ self ownerThatIsA: EnchantingConfigurationDialog) 
		ifNil: [ ^ self error: 'A configuration block should always be on an Enchanting Dialog!!' ]. 
	dlg freeUpName: previousName.
	newName _ dlg uniqueNameFor: (self currentName).
	dlg registerName: newName.
	self nameMorph stringExpression: newName.
	previousName _ newName.
! !

!ConfigurationBlockMorph methodsFor: 'event handling' stamp: 'CWB 1/1/2011 11:32'!
delete

	| dlg |

	(dlg _ self ownerThatIsA: EnchantingConfigurationDialog) ifNotNil: [
		dlg freeUpName: self currentName.
	].

	super delete.! !


!ConfigurationBlockMorph methodsFor: 'copying' stamp: 'CWB 9/22/2012 15:23'!
fullCopy

	| newSelf dlg newName |

	newSelf _ super fullCopy.
	(dlg _ self ownerThatIsA: EnchantingConfigurationDialog) ifNotNil: [
		newSelf currentName ifNotNil: [
			newName _ dlg uniqueNameFor: newSelf currentName.
			dlg registerName: newName.
			newSelf initialName: newName.
		].
	].
	^ newSelf.! !
