// Exported from Enchanting
package enchanting;

import java.io.OutputStream;
import java.io.PrintStream;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.game.Sprite;
import javax.microedition.lcdui.Font;
import lejos.nxt.LCD;

// This is based on LCDOutputStream, which was written by Lawrie Griffiths
public class StageOutputStream extends OutputStream {
	private int col = 0;
	private int line = 0;
	private Image image = null;
	private Graphics gfx = null;
	private boolean incLineBeforeNextOutput = false;

	public void setImage(Image i) {
		image = i;
		col = 0;
		line = 0;
		incLineBeforeNextOutput = false;
		
		// sanity check
		if (image.getWidth() != LCD.SCREEN_WIDTH || image.getHeight() != LCD.SCREEN_HEIGHT) {
			image = null;
		}
		if (image != null) {
			gfx = image.getGraphics();
		}
	}
	
		public void write(int c) {
		if (image == null) return;	
	
		if (incLineBeforeNextOutput) {
			incLine();
			incLineBeforeNextOutput = false;
		}
	
		char x = (char)(c & 0xFF);
		switch (x)
		{
			case '\t':
				col = col + 8 - col % 8; 
				break;
			case '\n': 
				incLineBeforeNextOutput = true;
			case '\r':
				col = 0;
				break;
			default:
				if (col >= LCD.DISPLAY_CHAR_WIDTH)
				{
					col = 0;
					incLine();
				}
				//LCD.drawChar(x, col++, line);
				gfx.drawChar(x, LCD.CELL_WIDTH * col++, LCD.CELL_HEIGHT * line, Graphics.LEFT | Graphics.TOP);
		}
		
	}

	private void incLine() {
		if (image == null) return;	
		byte[] buffer = image.getData();
		
		if (line < LCD.DISPLAY_CHAR_DEPTH - 1)
			line++;
		else {
			//LCD.scroll();

	        LCD.bitBlt(
	        	buffer, LCD.SCREEN_WIDTH, LCD.SCREEN_HEIGHT,  	// src, sw, sh
	        	0, LCD.CELL_HEIGHT,								// sx, sy
	        	buffer, LCD.SCREEN_WIDTH, LCD.SCREEN_HEIGHT, 	// dst, dw, dh
                0, 0, LCD.SCREEN_WIDTH,							// dx, dy, w
                LCD.SCREEN_HEIGHT - LCD.CELL_HEIGHT,			// w, h
                LCD.ROP_COPY);									// rop
                
    	    LCD.bitBlt(
    	    	null, LCD.SCREEN_WIDTH, LCD.SCREEN_HEIGHT, 		// src, sw, sh
    	    	0, 0, 											// sx, sy
	        	buffer, LCD.SCREEN_WIDTH, LCD.SCREEN_HEIGHT, 	// dst, dw, dh
    	    	0, LCD.SCREEN_HEIGHT - LCD.CELL_HEIGHT,			// dx, dy
                LCD.SCREEN_WIDTH, LCD.CELL_HEIGHT, 				// w, h
                LCD.ROP_CLEAR);									// rop
		}
	}
}
